package main;

import static main.DatalogGrammar.V.*;
import static main.DatalogGrammar.Σ.*;
import static fling.grammars.api.BNFAPI.bnf;
import static fling.internal.grammar.sententials.Notation.*;

import java.io.IOException;
import java.nio.file.*;
import java.util.*;
import java.util.Map.Entry;
import java.util.function.Supplier;

import com.google.googlejavaformat.java.Formatter;
import com.google.googlejavaformat.java.FormatterException;

import fling.*;
import fling.BNF;
import fling.adapters.JavaMediator;

/**
 * Datalog language.
 * 
 * @author Yossi Gil
 */
public class DatalogGrammar {
  /** Set of terminals, i.e., method names of generated fluent API. */
  public enum Σ implements fling.Terminal {
    infer, fact, query, of, and, when, always, v, l
  }

  /**
   * Set of non-terminals, i.e., abstract concepts of fluent API; these names
   * will be translated into names of classes of abstract syntax tree that Fling
   * generates, i.e., this AST will have class {@link Program} which will have a
   * list of {@link Statement}, etc.
   */
  public enum V implements fling.Variable {
    Program, Statement, Rule, Query, Fact, Bodyless, WithBody, //
    RuleHead, RuleBody, FirstClause, AdditionalClause, Term
  }

  /**
   * Short name of {@link String}.class, used to specify the type of parameters
   * to fluent API methods in grammar specification.
   */
  private static final Class<String> S = String.class;
  /**
   * Datalog's grammar in Backus-Naur form.
   */
  // @formatter:off
  public static final BNF bnf = bnf(). //
      start(Program). // This is the start symbol
      derive(Program).to(oneOrMore(Statement)). //
      /* Defines the rule Program ::= Statement. The 'Program' AST node type
       * generated by Fling includes a field of type List containing the
       * Statements. */
      specialize(Statement).into(Fact, Rule, Query). //
      /* Defines the rule Statement ::= Fact | Rule | Query, but also defines
       * that classes {@link Fact}, {@link Rule} and {@link Query} extend class
       * {@link Statement}. */
      derive(Fact).to(fact.with(S), of.many(S)). // Fact ::= fact(S) of(S*)
      derive(Query).to(query.with(S), of.many(Term)). // Query ::= query(S) of(Term*)
      specialize(Rule).into(Bodyless, WithBody). // Rule ::= Bodyless | WithBody
      derive(Bodyless).to(always.with(S), of.many(Term)). // Bodyless ::= always(S) of(Term*)
      derive(WithBody).to(RuleHead, RuleBody). // WithBody ::= RuleHead RuleBody
      derive(RuleHead).to(infer.with(S), of.many(Term)). // RuleHead ::= infer(S) of(Term*)
      derive(RuleBody).to(FirstClause, noneOrMore(AdditionalClause)). // RuleBody ::= FirstClause AdditionalClause*
      derive(FirstClause).to(when.with(S), of.many(Term)). // FirstClause ::= when(S) of(Term*)
      derive(AdditionalClause).to(and.with(S), of.many(Term)). // AdditionalClause ::= and(S) of(Term*)
      derive(Term).to(l.with(S)).or(v.with(S)). // Term ::= l(S) | v(S)
      build();
  // @formatter:on
  /**
   * The {@link JavaMediator} responsible for compiling the Java Datalog API/AST
   * types/AST run-time compiler.
   */
  public static final JavaMediator jm = new JavaMediator(//
      bnf, // Use this BNF as language specification.
      "main", // Name of package in which output will reside.
      "Datalog", // Name of generated API.
      Σ.class // Language terminals enum.
  );

  /**
   * Prints the Datalog API/AST types/AST run-time compiler to corresponding
   * files.
   */
  public static void main(String[] args) throws IOException, FormatterException {
    Map<String, String> files = ((Supplier<Map<String, String>>) () -> {
      final Map<String, String> $ = new LinkedHashMap<>();
      $.put("Datalog", DatalogGrammar.jm.apiClass);
      $.put("DatalogAST", DatalogGrammar.jm.astClass);
      $.put("DatalogCompiler", DatalogGrammar.jm.astCompilerClass);
      return $;
    }).get();
    String PATH = "./src/main/";
    System.out.println("project path: " + PATH);
    final Path outputFolder = Paths.get(PATH);
    if (!Files.exists(outputFolder)) {
      Files.createDirectory(outputFolder);
      System.out.println("directory " + PATH + " created successfully");
    }
    final Formatter formatter = new Formatter();
    for (final Entry<String, String> file : files.entrySet()) {
      final Path filePath = Paths.get(PATH + file.getKey() + ".java");
      if (Files.exists(filePath))
        Files.delete(filePath);
      Files.write(filePath, Collections.singleton(formatter.formatSource(file.getValue())), StandardOpenOption.CREATE,
          StandardOpenOption.WRITE);
      System.out.println("file " + file.getKey() + ".java written successfully.");
    }
  }
}
